@@ -27,7 +27,9 @@ import math as math_module
 import shutil
 import typing
 import html
+import warnings
 from .latex2svg import latex2svg
+from .helpers import heading_name_to_id_value
 
 MODULE_PATH = os.path.join(*os.path.split(__file__)[:-1])
 DEBUG = False  # weather to print debug information
@@ -154,9 +156,7 @@ def find_and_replace_formulas_in_markdown(md: str, support_formulas=True):
 
 def header_name_to_link_to_header(header_name: str) -> str:
     """Takes the name of a header and creates a link to said header in markdown format."""
-    pattern = re.compile('[\W_]+')
-    id_from_title = "-".join(
-        [pattern.sub('', html.unescape(word).strip()).lower().replace("-", "") for word in header_name.strip().split()])
+    id_from_title = heading_name_to_id_value(header_name)
     link = "[" + header_name.strip() + "](#" + id_from_title + ")"
     # We do NOT add `#user-content` here! This is on purpose since this is done at another place later.
     return link
@@ -549,7 +549,7 @@ def main(md_origin, origin_type="file", website_root=None, destination=None, ima
          output_name="<name>.html", output_pdf=None, style_pdf="True", footer=None, math="True",
          formulas_supporting_darkreader=False, extra_css=None,
          core_converter: typing.Union[str, typing.Callable] = markdown_to_html_via_github_api,
-         compress_images=False, enable_image_downloading=True, box_width=None, toc=False):
+         compress_images=False, enable_image_downloading=True, box_width=None, toc=False, dont_make_images_links=False):
     # set all to defaults:
     style_pdf = str2bool(style_pdf)
     math = str2bool(math)
@@ -558,7 +558,9 @@ def main(md_origin, origin_type="file", website_root=None, destination=None, ima
         website_root = ""
     if destination is None:
         destination = ""
-    if image_paths is None:
+    if not image_paths:
+        if image_paths == "":
+            enable_image_downloading = False
         image_paths = destination + (os.sep if destination else "") + "images"
     if css_paths is None:
         css_paths = "github-markdown-css"
@@ -634,6 +636,15 @@ def main(md_origin, origin_type="file", website_root=None, destination=None, ima
     if DEBUG:
         print("\n------------\nHtml content:\n------------\n\n", html_content)
 
+    # remove links around images if requested:
+    if dont_make_images_links:
+        html_bs4 = BeautifulSoup(html_content, "html.parser")
+        for img_bs4 in html_bs4.find_all("img"):
+            if img_bs4.parent.name == "a":
+                if img_bs4.parent["href"] == img_bs4["src"]:
+                    img_bs4.parent.unwrap()
+        html_content = html_bs4.__str__()
+
     # re-insert formulas in html, this time as proper svg images:
     html_content = find_and_render_formulas_in_html(html_content, formula_mapper, special_chars_in_code_blocks)
 
@@ -801,7 +812,7 @@ def main(md_origin, origin_type="file", website_root=None, destination=None, ima
                 if compression_information["srcset"] and not width:
                     srcset = compression_information["srcset"]
                     srcset.sort()
-                    srcset = list(filter((lambda x: x < full_image.width), srcset))
+                    srcset = [x for x in srcset if x < full_image.width]
                     srcset.append(full_image.width)
                     # Create all the compressed images and a srcset-attribute for them:
                     srcset_attribute = str()
@@ -841,9 +852,10 @@ def main(md_origin, origin_type="file", website_root=None, destination=None, ima
                     height = img_object.height
                 max_height_css_information = "max-height: " + str(height) + "px;"
                 if img_soup_representation.has_attr("style"):
-                    img_soup_representation["style"] = (
-                        img_soup_representation["style"].strip().rstrip(";") + "; " + max_height_css_information
-                    )
+                    if ";max-height:" not in ";" + img_soup_representation["style"].replace(" ", ""):
+                        img_soup_representation["style"] = (
+                            img_soup_representation["style"].strip().rstrip(";") + "; " + max_height_css_information
+                        )
                 else:
                     img_soup_representation["style"] = max_height_css_information
             # Change src/href tags to ensure we reference the right image:
@@ -873,11 +885,14 @@ def main(md_origin, origin_type="file", website_root=None, destination=None, ima
         print("\n------------\nHtml with image links:\n------------\n\n", html_rendered)
 
     # ensure <a href="#fufu"> gets converted to <a href="#user-content-fufu">:
+    contains_file_internal_links = False
     html_soup = BeautifulSoup(html_rendered, 'html.parser')
     for link_soup_representation in html_soup.find_all("a"):
         link_location = link_soup_representation.get("href")
         if link_location.startswith("#"):
             link_location = "#user-content-" + link_location[1:]
+            if not (link_soup_representation.has_attr("class") and link_soup_representation["class"] == ["anchor"]):
+                contains_file_internal_links = True
         link_soup_representation["href"] = link_location
     html_rendered = html_soup.__str__()
 
@@ -891,26 +906,19 @@ def main(md_origin, origin_type="file", website_root=None, destination=None, ima
     #       link_within_header.append(BeautifulSoup(GITHUB_LINK_ANCHOR, 'html.parser').find("svg"))
     html_rendered = html_soup.__str__()
 
-    # turns out to not be necessary (
-    # un-escape all `href`s und `src`s (they are all html-escaped, for some reason):
-    # html_soup = BeautifulSoup(html_rendered, 'html.parser')
-    # for tag in ("a", "img"):
-    #     for link_or_img_soup_representation in html_soup.find_all(tag):
-    #         for attr_name in ("src", "href", "data-canonical-src"):
-    #             if link_or_img_soup_representation.has_attr(attr_name):
-    #                 print("FOO&amp;Bar", link_or_img_soup_representation[attr_name][10:],
-    #                       html.unescape(link_or_img_soup_representation[attr_name])[10:])
-    #                 link_or_img_soup_representation[attr_name] = html.unescape(
-    #                     link_or_img_soup_representation[attr_name])
-    # html_rendered = html_soup.__str__()
-
     if DEBUG:
         print("\n------------\nHtml with fixed internal links:\n------------\n\n", html_rendered)
 
     # ensure we have the css and the code navigation banner where we want it to be (anyone knows what this is for?):
     with open_local("github-css.min.css", "r") as from_f:
+        github_min_css = from_f.read()
+        if core_converter in ("OFFLINE+", "OFFLINE"):
+            # add syntax highlighting css if we use OFFLINE or OFFLINE+ for conversion:
+            from pygments.formatters import html as pygments_html
+            github_min_css += pygments_html.HtmlFormatter().get_style_defs('.highlight')\
+                .replace("\n", "").replace(";", " !important;").replace(" }", " !important }")
         with open(os.path.join(abs_css_paths, "github-css.css"), "w") as to_f:
-            to_f.write(from_f.read())
+            to_f.write(github_min_css)
     with open_local("code-navigation-banner-illo.svg", "r") as from_f:
         with open(os.path.join(abs_css_paths, "code-navigation-banner-illo.svg"), "w") as to_f:
             to_f.write(from_f.read())
@@ -934,17 +942,22 @@ def main(md_origin, origin_type="file", website_root=None, destination=None, ima
 Unfortunately, you need to have pdfkit installed to save as pdf. Find out how to install it here:
 https://pypi.org/project/pdfkit/""")
         # ensure the image srcs are absolute file paths:
-        # ToDo: Completely disable pdf conversion on windows machines, since wkhtmltopdf won't run on them anyways.
         html_soup_representation = BeautifulSoup(html_rendered, "html.parser")
-        tags_with_links = html_soup_representation.find_all(lambda tag: tag.has_attr("src"))
-        for tag_with_link in tags_with_links:
-            link = tag_with_link["src"]
-            if link.startswith("https://") or link.startswith("http://"):
-                pass
-            else:
-                directory_to_link_to = abs_website_root if link.startswith("/") else abs_destination
-                abs_path = urllib.parse.quote(os.path.join(os.getcwd(), directory_to_link_to.lstrip("/"), link))
-                tag_with_link["src"] = "file://" + abs_path
+        for filter, attr in (((lambda tag: tag.has_attr("src")), "src"), ("link", "href")):
+            for tag_with_link in html_soup_representation.find_all(filter):
+                link = tag_with_link[attr]
+                if "://" in link:
+                    pass
+                else:
+                    directory_to_link_to = abs_website_root if link.startswith("/") else abs_destination
+                    if directory_to_link_to == "./":
+                        directory_to_link_to = ""
+                    cwd = str(os.getcwd())
+                    abs_path = urllib.parse.quote(os.path.join(cwd, directory_to_link_to.strip("/"), link.strip("/")))
+                    if link.startswith("/") and not enable_image_downloading and attr == "src":
+                        # link is already absolute
+                        abs_path = link
+                    tag_with_link[attr] = "file://" + abs_path
         html_rendered = html_soup_representation.__str__()
 
         # Remove links that link to something that clearly lies on the disk or is a relative link, since these won't
@@ -961,7 +974,6 @@ https://pypi.org/project/pdfkit/""")
                     link.parent.append(child)
                 link.decompose()
         html_rendered = html_soup_representation.__str__()
-        # ToDo: Make sure that anchor links work correctly... but that's a task for another day.
 
         # remove the css if we want to save it without css:
         if not style_pdf:
@@ -973,6 +985,24 @@ https://pypi.org/project/pdfkit/""")
         else:
             output_pdf = output_pdf.replace("<name>", file_name_origin)
 
+        # check which version of wkhtmltopdf we have installed
+        version_str = str(subprocess.check_output(["wkhtmltopdf", "-V"]), encoding="UTF-8").strip()
+        if version_str.startswith("wkhtmltopdf "):
+            version_str = version_str.split(" ", 1)[1]
+        version_number = [int(i) for i in version_str.split()[0].split(".")]
+        given_version_as_number = 10**12 * version_number[0] + 10**6 * version_number[1] + version_number[2]
+        ideal_version_as_number = 10**12 * 0 + 10**6 * 12 + 6
+        options = {'quiet': ''}
+        if given_version_as_number >= ideal_version_as_number:
+            options['enable-local-file-access'] = ''
+            # ^ see https://github.com/wkhtmltopdf/wkhtmltopdf/issues/2660#issuecomment-663063752
+        if contains_file_internal_links and "(with patched qt)" not in version_str:
+            warning_text = ("Your file contains internal links, but your version of wkhtmltopdf (version \""
+                            + version_str + "\") does not support using these,\n\tsince you need a version with the qt"
+                            + " patches to use internal links within your file.\n\tYou can download these from"
+                            + " https://wkhtmltopdf.org/downloads.html")
+            warnings.warn(warning_text, Warning)
+
         # Finally convert it:
         # (this saving and then converting ensures we don't convert links like https:foo to
         # absolute_path_to_file_location/https://foo)
@@ -981,7 +1011,7 @@ https://pypi.org/project/pdfkit/""")
         pdfkit.from_file(
             os.path.join(abs_destination, output_pdf + ".html"),
             os.path.join(destination, output_pdf),
-            options=dict() if DEBUG else {'quiet': ''},
+            options=dict() if DEBUG else options,
         )
         os.remove(os.path.join(abs_destination, output_pdf + ".html"))
 
@@ -1053,9 +1083,10 @@ if __name__ == "__main__":
     parser.add_argument('-d', '--destination', nargs="+", action=FuseInputString, help="""
     Where to store the generated html. This path is relative to --website-root. Defaults to "".""")
 
-    parser.add_argument('-i', '--image-paths', nargs="+", action=FuseInputString, help="""
+    parser.add_argument('-i', '--image-paths', nargs="*", const="", action=FuseInputString, help="""
     Where to store the images needed or generated for the html. This path is relative to website-root. Defaults to the
-    "images"-folder within the destination folder.""")
+    "images"-folder within the destination folder.
+    Leave this option empty to completely disable image caching/downloading and leave all image links unmodified.""")
 
     parser.add_argument('-c', '--css-paths', nargs="+", action=FuseInputString, help="""
     Where to store the css needed for the html (as a path relative to the website root). Defaults to the
@@ -1070,7 +1101,9 @@ if __name__ == "__main__":
     parser.add_argument('-p', '--output-pdf', nargs="+", action=FuseInputString, help="""
     If set, the file will also be saved as a pdf file in the same directory as the html file, using pdfkit, a python
     library which will also need to be installed for this to work. You may use the <name> variable in this value like
-    you did in --output-name.""")
+    you did in --output-name.
+    Do not use this with the -c option if the input of the -c option is not trusted; execution of malicious code might
+    be the consequence otherwise!!""")
 
     parser.add_argument('-s', '--style-pdf', default="true", help="""
     If set to false, the generated pdf (only relevant if you use --output-pdf) will not be styled using github's css.
